from contextlib import contextmanager
from http.server import BaseHTTPRequestHandler, HTTPServer
import json
import logging
import threading
import urllib



class CallbackRequestHandler(BaseHTTPRequestHandler):

    def do_GET(self):

        # print("GET request\nPath: {}\nHeaders:\n{}".format(str(self.path), str(self.headers))) # log.INFO

        split_path = self.path.split('?')

        if len(split_path) != 2:
            raise Exception("No parameter query in HTTP call.")

        parsed_parms = urllib.parse.parse_qs(split_path[1])

        if not 'code' in parsed_parms:
            raise Exception("No 'code' field in query of HTTP call.")
        
        # if 'Content-Length' in self.headers:
        #     content_len = int(self.headers.get('Content-Length'))
        #     post_body = self.rfile.read(content_len)
        #     print('Post Body:') # log.INFO
        #     pprint(dict(urllib.parse.parse_qsl(post_body.decode()))) # log.INFO

        self.send_response(200)
        self.send_header('Content-type', 'text/json')
        self.end_headers()

        # self.wfile.write("GET request for {}".format(self.path).encode('utf-8'))
        self.wfile.write(json.dumps({ k:v[0] for k,v in parsed_parms.items() }).encode('utf-8'))

    def do_POST(self):

        logging.warning("SERVER POSTED TO KILL.")
        self.send_response(200)
        self.send_header('Content-type', 'text/json')
        self.end_headers()

        self.wfile.write(json.dumps({"status":"killed"}).encode('utf-8'))



class Oath2CallbackServer(threading.Thread):
    def __init__(self, event=None, address='', port=8080):
        threading.Thread.__init__(self)
        self.event = event
        self.address=address
        self.port = int(port)

    def run(self):
        logging.debug("Starting httpd...\n") # log.INFO

        server_address = (self.address, self.port)
        httpd = HTTPServer(server_address, CallbackRequestHandler)

        if self.event: self.event.set()

        try:
            # handles incoming request once
            httpd.handle_request()
            logging.debug("REQUEST HANDLED") # log.INFO
        except KeyboardInterrupt:
            pass
        finally:
            httpd.server_close()
            logging.debug("Stopping httpd...\n") # log.INFO



@contextmanager
def local_callback_server(callback_url):

    import requests
    import threading

    # Grab the port of the callback_url
    # server_address = re.split(r'(.+):(.+)[/]', callback_url)
    # port = server_address[2]

    logging.debug('Start callback server') # log.INFO

    evt = threading.Event()
    callback_handler = Oath2CallbackServer(evt)#, port=port)
    callback_handler.start()
    # Wait until the server is loaded up before kicking off the callback api
    evt.wait()
    
    try:
        yield
    except:
        # Tell the Oath2CallbackServer to quit.
        requests.post(callback_url)
        raise
    finally:
        callback_handler.join()



from html.parser import HTMLParser
import re


class CSRFParser(HTMLParser):

    def __init__(self):
        super().__init__()
        self.secret = None
    
    def handle_starttag(self, tag, attrs):
        '''Looking for {"name":"csrf_token", "type":"hidden", "value": "####"}'''
        if tag == "input":
            secrets = dict(attrs)
            if 'name' in secrets and secrets['name'] == 'csrf_token':
                self.secret = secrets['value']

    def get_secret(self):
        return self.secret


class UserParser(HTMLParser):

    def __init__(self, username):
        super().__init__()
        self.username = username
        self.check_data = False
        self.user_id = None
        self.url_query = None

    def handle_starttag(self, tag, attrs):
        if tag == "a":
            a = dict(attrs)
            if 'href' in a and 'selecteduser' in a['href']:
                self.url_query = a['href']
                self.check_data = True

    def handle_data(self, data):
        if self.check_data:
            if data == self.username:
                self.user_id = re.search(r'selecteduser=(\d+)', self.url_query).groups()[0]
            self.check_data = False

    def get_user_id(self):
        return self.user_id



from datetime import datetime
from datetime import timedelta
from functools import partialmethod
import json
import logging
from math import floor
from threading import Event
from threading import Lock

from dateutil.parser import parse
from oauthlib.oauth2.rfc6749.errors import MissingTokenError
from requests import Response
from requests_oauthlib import OAuth2Session

from .codes import MeasureType, SleepState
from .exceptions import AuthenticationFailedException
from .exceptions import raise_for_status as withings_raise_for_status



def restructure_token(response: Response):
    """
    Withings returns token contents in 'body' of json dict.
    Pull those contents out to the first level of a token dict.
    """

    try:
        resp = response.json()
    except json.decoder.JSONDecodeError as e:
        logging.error(response.text)
        raise
        # alternatively, the unmodified response could be returned here

    new_token = {}

    status = resp.get('status', None)
    if status:
        # If no error, status was set to 0 and this block won't run
        new_token['error'] = status

    body = resp.get('body', None)
    if body:
        new_token.update(body)

    response._content = json.dumps(new_token).encode("UTF-8")

    return response


class WithingsOath2Client:
    '''
    access_token is valid for three hours.
    Use refresh_token to get a new access_token after it expires.
    '''

    def __init__(self, client_id, client_secret, callback_url,
                 auth_code=None, access_token=None, 
                 refresh_token=None, expires_at=None,
                 token_updater=None):

        if not auth_code and not (access_token and refresh_token and expires_at):
            raise Exception("Either auth_code, or access_token & refresh_token & expires_at must be given.")
        elif auth_code and access_token and refresh_token:
            raise Exception("Only auth_code, or access_token & refresh_token must be given.")

        self.client_id = client_id
        self.client_secret = client_secret

        self.auth_code = auth_code
        self.callback_url = callback_url
        self.access_token = access_token
        self.refresh_token = refresh_token

        # Handle setting expires_at value to float timestamp
        try:
            self.expires_at = floor(float(expires_at))
        except ValueError as err:
            if 'could not convert string to float' in str(err):
                self.expires_at = parse(expires_at).timestamp()
            else:
                raise
        except TypeError as err:
            self.expires_at = expires_at

        self._token_updater = token_updater or (lambda x: x)

        token = {}
        if access_token: token['access_token'] = access_token
        if refresh_token: token['refresh_token'] = refresh_token
        if expires_at: token['expires_at'] = self.expires_at

        self.refresh_lock = Lock()
        self.refresh_event = Event()

        self.session = OAuth2Session(client_id=self.client_id,
                                     auto_refresh_url='https://wbsapi.withings.net/v2/oauth2',
                                     auto_refresh_kwargs={
                                         "action": "requesttoken",
                                         "client_id": self.client_id,
                                         "client_secret": self.client_secret,
                                     },
                                     token=token,
                                     redirect_uri=self.callback_url,
                                     scope='user.info,user.metrics,user.activity',
                                     token_updater=self._token_updater)

        # Hooks to restructure token dict upon access & refresh of session tokens
        self.session.register_compliance_hook("access_token_response", restructure_token)
        self.session.register_compliance_hook("refresh_token_response", restructure_token)


    def fetch_access_token(self, refresh=False):
        '''After this point, the contents of the response dict are saved in session
        as an oauthlib.OAuth2Token object.
        session.token accesses the object, while session.access_token gets just
        the string of latest access token.
        '''

        if not self.refresh_lock.acquire(blocking=False):
            logging.info("WAITING ON ANOTHER THREAD TO REFRESH ACCESS TOKEN.")
            self.refresh_event.wait()
            return
        # self.refresh_event.clear()

        try:
            fetch_token_url = 'https://wbsapi.withings.net/v2/oauth2'

            if not refresh:
                logging.info("FETCHING ACCESS TICKET")

                response = self.session.fetch_token(fetch_token_url,
                                                    include_client_id=True,
                                                    client_secret=self.client_secret,
                                                    code=self.auth_code,
                                                    action="requesttoken")

                # After this point, the contents of the response dict are saved in session
                # as an oauthlib.OAuth2Token object.
                # session.token accesses the object, while session.access_token gets just
                # the string of latest access token.

            else:
                logging.info("REFRESHING ACCESS TICKET")
                response = self.session.refresh_token(fetch_token_url)

            # Call token_updater callback function to save off new creds
            if self._token_updater:
                logging.debug("Handling new creds with token_updater callback")
                self._token_updater(response)

            # Take in new auth details
            self.access_token = response['access_token']
            self.refresh_token = response['refresh_token']
            self.expires_at= floor(response['expires_at'])
            
            logging.info("TOKEN AUTHENTICATED")

        except MissingTokenError as err:
            # Probably ALWAYS means there wasn't an Access Token returned in the response
            # due to some issue in the provided parameters of the request
            logging.error("MissingTokenError raised.")
            raise

        finally:
            self.refresh_event.set()
            self.refresh_event.clear()
            self.refresh_lock.release()

        return response


    # TODO: implement backoff once proper exceptions are raised
    def _request(self, method, url, **kwargs):

        if not self.session.token:
            logging.debug("NO TOKEN YET")
            self.fetch_access_token()

        # if not self.expires_at:# and self.expires_at - datetime.now() < 0:
        if self.expires_at \
                and datetime.fromtimestamp(self.expires_at) < datetime.now():
            logging.debug("Withings API token has expired")
            self.fetch_access_token(refresh=True)

        try:
            response = self.session.request(method, url, **kwargs)
            response.raise_for_status()
            withings_raise_for_status(response)

        # except oauthlib.oath2.TokenExpiredError as e:
        # except requests.exceptions.HTTPError as err:
        except AuthenticationFailedException as err:

            content = json.loads(response.content.decode('utf8'))

            if content['status'] == 401:
                if 'The access token provided' in content['error']:
                    self.fetch_access_token(refresh=True)
                    response = self.session.request(method, url, **kwargs)
                else:
                    raise

        return response


    get = partialmethod(_request, 'GET')
    post = partialmethod(_request, 'POST')



class Withings:
    # access_token is valid for three hours.
    # Use refresh_token to get a new access_token after it expires.


    def __init__(self, client_id, client_secret, callback_url,
                 auth_code=None, access_token=None,
                 refresh_token=None, expires_at=None,
                 token_updater_cb=None):

        self.client = WithingsOath2Client(client_id, client_secret, callback_url, 
                                          auth_code, access_token, 
                                          refresh_token, expires_at,
                                          token_updater_cb)


        # TODO: deprecate this
        self.headers = { "Authorization": "Bearer {}".format("HAH!") }#self.access_token) }


    def get_devices(self):
        """
        {'body': {'devices': [{'battery': 'high',
                       'deviceid': '...',
                       'hash_deviceid': '...',
                       'last_session_date': ...,
                       'model': 'Aura Sensor V2',
                       'model_id': 63,
                       'timezone': 'America/Chicago',
                       'type': 'Sleep Monitor'},
                      {'battery': 'medium',
                       'deviceid': '...',
                       'hash_deviceid': '...',
                       'last_session_date': ...,
                       'model': 'Body+',
                       'model_id': 5,
                       'timezone': 'America/Chicago',
                       'type': 'Scale'}]},
        'status': 0}
        """

        parms = { "action": "getdevice" }

        response = self.client.get('https://wbsapi.withings.net/v2/user',
                                   headers=self.headers,
                                   params=parms)

        results = response.json()

        if 'body' not in response and 'devices' not in results['body']:
            logging.error('ERROR:','No body/devices in results.')
            logging.error(response.text)
            return []

        return results['body']['devices']


    def get_measurements(self, type_name):
        """Fetches measurements.

        Parameters
        ----------
        type_name : str, int, MeasureType
            The string value, or integer code of the MeasureType, or
            even the MeasureType enum itself.
        """

        parms = { "action": "getmeas" }

        hdrs = self.headers

        if type_name:
            hdrs['type'] = str(MeasureType(type_name).code)

        response = self.client.get('https://wbsapi.withings.net/measure',
                                   headers=hdrs,
                                   params=parms)

        measurements = response.json()

        # Replace type-codes with type-names
        for record in measurements['body']['measuregrps']:
            for measure in record['measures']:
                measure['type'] = MeasureType(measure['type']).value
                # 'algo', 'fm', 'type', 'unit', 'value'

        return measurements['body']['measuregrps']


    def get_sleep_data(self, 
                       start_date=datetime(year=2019, month=5, day=16), 
                       end_date=datetime.now() + timedelta(days=1)):
        '''
        First viable startdate is default.
        Tomorrow is default enddate, to ensure latest night's data.
        If end_date > 24hrs after start_date, the API will instead use 24hrs after specified start_date.
        '''

        parms = {
            "action": "get",
            "startdate": start_date.strftime('%s'),
            "enddate": end_date.strftime('%s'),
            #"data_fields": 'hr,rr,snoring',
        }

        logging.debug('Fetching data between {} and {}'.format(start_date.date(), end_date.date()))

        response = self.client.get('https://wbsapi.withings.net/v2/sleep', params=parms)
        
        results = response.json()
        
        if 'body' not in results and 'series' not in results['body']:
            logging.error('ERROR:','No body/series in results.')
            logging.error(response.text)
            # '{"status":601,"body":{"wait_seconds":68},"error":"Too Many Requests"}'
            return []

        for sleep in results['body']['series']:
            sleep['state'] = SleepState(sleep['state']).value
        
        return results['body']['series']


    def get_sleep_detail_data(self, 
                       start_date=None, 
                       end_date=None,
                       last_update_date=None):
        '''
        If 'more'=True and 'offset'>0 in result['body'], then offset value is the 
        number of returned data from the current request, and there is more data available.
        Seems like this API returns MAX 300 records per request.
        '''

        parms = {
            "action": "getsummary",
            #"data_fields": 'hr,rr,snoring',
        }

        if (start_date and end_date) and not last_update_date:
            parms["startdate"] = start_date.strftime('%s')
            parms["enddate"] = end_date.strftime('%s')

        elif last_update_date and (not start_date and not end_date):
            parms["lastupdate"] = last_update_date.strftime('%s')

        else:
            raise Exception("Either last_update_date or start & end dates should be provided.")

        response = self.client.get('https://wbsapi.withings.net/v2/sleep',
                                   headers=self.headers,
                                   params=parms)
        
        results = response.json()
        
        # if 'series' not in results['body']:
        #     logging.error('ERROR:','No body/series in results.')
        #     logging.error(response.text)
        #     return []

        return results['body']['series'], response.json()['body']['more']


    def subscribe(self, callback_url, noti_category):

        response = self.client.post(
            'https://wbsapi.withings.net/notify',
            params={
                'action': 'subscribe',
                'callbackurl': callback_url,
                'appli': noti_category,
            })

        return response

    def unsubscribe(self, callback_url, noti_category):

        response = self.client.post(
            'https://wbsapi.withings.net/notify',
            params={
                'action': 'revoke',
                'callbackurl': callback_url,
                'appli': noti_category,
            })

        return response



from json.decoder import JSONDecodeError
import logging
import re
import requests
import urllib

from .callback import local_callback_server
from ..exceptions import MismatchingRedirectURIError
from .parser import CSRFParser
from .parser import UserParser



def get_url_host(url_path):
    parsed_uri = urllib.parse.urlparse(url_path)
    host = parsed_uri.netloc
    return host


def get_url_params(url_path):
    split_path = url_path.split('?')

    if len(split_path) != 2:
            raise Exception("No parameter query in url.")

    parsed_parms = urllib.parse.parse_qs(split_path[1])

    return { k:v[0] for k,v in parsed_parms.items() }


# TODO: add exceptions from responses
# oauthlib.oauth2.rfc6749.errors
#   InvalidScopeError

class WithingsAUTH:

    def __init__(self, client_id, callback_url, state='ABCDEFG'):

        self._session = requests.Session()

        self.callback_url = callback_url
        # self.client_id = client_id

        self.headers = {
            'Host': 'account.withings.com',
            'Connection': 'keep-alive',
            'Content-Type': 'application/x-www-form-urlencoded',
        }

        self.parms = {
            'response_type': 'code',
            'client_id': client_id,
            'state': state,
            # Scale - user.metrics
            # Sleep - user.activity
            'scope': 'user.info,user.metrics,user.activity',
            'redirect_uri': callback_url, #redirect_uri,
            'b': 'authorize2',
        }


    def _call(self, verb: str, route: str, data=None, **kwargs):

        if data:
            payload = urllib.parse.urlencode(data)

        response = self._session.request(
            method=verb,
            url="https://account.withings.com/oauth2_user/{}".format(route),
            params=urllib.parse.urlencode(self.parms),
            headers=self.headers,
            data=data,
            **kwargs)

        return response


    def _get_csrf_token(self, username=None, **kwargs):

        response = self._call('get', 'account_login', **kwargs)

        # Get ending route of url to check where we're at
        route = re.search(r'\/([\w_]+)\?', response.url).groups()[0]

        # if there are multiple users in this account, there will be a
        # select-user page interjected here. Use the provided username to
        # get the appropriate user_id value
        if route == 'user_select':
            user_parser = UserParser(username)
            user_parser.feed(response.text)
            user_id = user_parser.get_user_id()
            self.parms['selecteduser'] = user_id
            response = self._call('get', 'account_login')
            logging.debug('User {} selected'.format(username))

        # if routed response comes up with 'selecteduser' param, save value
        elif 'selecteduser' in response.url:
            query = get_url_params(response.url)
            self.parms['selecteduser'] = query['selecteduser']
            logging.debug('selecteduser acquired.')

        # Extract csrf_token value from html element
        parser = CSRFParser()
        parser.feed(response.text)
        csrf_token = parser.get_secret()

        if not csrf_token:
            resp = response.json()

            if 'redirect_uri_mismatch' in resp['errors'][0]['message']:
                raise MismatchingRedirectURIError(self.callback_url)
            else:
                logging.error(resp)
                raise Exception("No CSRF Token element found on HTML page.")

        logging.debug('CSRF_TOKEN acquired.')

        return csrf_token


    def _sign_in(self, csrf_token: str, account_email, account_password):
        
        body = {
            'email': account_email,
            'password': account_password,
            'is_admin': 'f',
            'csrf_token': csrf_token,
        }

        response = self._call("post", "account_login", data=body)

        # check response.status_code for failed logins
        if 'session_key' not in self._session.cookies:
            logging.debug(response.url)
            raise Exception("""Sign-in error: no session_key provided. 
                Was there already too many failed attempts?""")

        logging.debug('SESSION_KEY Cookie acquired')

        return self._session.cookies['session_key']


    def _authorize(self, csrf_token, **kwargs):
        """
        For API Keys and such, it's recommended to pass 'auth' into kwargs as a requests auth object,
        rather than replacing headers.
        """

        body = {
            'authorized': 1,
            'csrf_token': csrf_token,
        }

        # Interject updating headers in-between redirect to ensure proper headers are sent to callback server
        first_response = self._call("post", "authorize2", data=body, allow_redirects=False, **kwargs)
        redirect_request = first_response._next
        # Withings leaves this as their own 'account.withings.com', potentially breaking the redirect with AWS
        # redirect_request.headers['Host'] = '*.execute-api.*.amazonaws.com'
        redirect_request.headers['Host'] = get_url_host(self.callback_url)

        response = self._session.send(redirect_request)

        try:
            if response.status_code == 403:
                from pprint import pprint
                pprint(response.__dict__)
                raise Exception('403 Response Status')
            else:
                result = response.json()
                code = result['code']
        except JSONDecodeError as e:
            logging.error(response.text)
            raise

        logging.debug('Authorization Code acquired: {}'.format(response.json()))

        return code
        #except (ConnectionResetError, ProtocolError, requests.exceptions.ConnectionError):


    def authorize(self, account_email, account_password, username=None, callback_server_gen=local_callback_server, **kwargs):
        '''username may not be necessary, if there is only a single user in the account.'''

        # GET:  grab csrf_token from a hidden element in the html on the ACCOUNT_LOGIN page
        csrf_token = self._get_csrf_token()

        # POST: attach that csrf to body along with creds to post to basically that same URL
        self._sign_in(csrf_token, account_email, account_password)

        logging.debug("SESSION KEY: {}".format(self._session.cookies['session_key']))

        # GET: grab new csrf_token, but with the session_id cookie included from the previous call
        #       to be taken to the 'Allow this app?' page
        # Also grabs the selecteduser ID
        new_csrf_token = self._get_csrf_token(username)

        with callback_server_gen(self.callback_url):
            # Request Authorization Code. Pulls from own server after redirect.
            # Posts to the AUTHORIZE2 page, which forwards to the encompassing CallbackServer, with a 'code' query 
            # parameter, which the CallbackServer reads and returns back to this authorize response.
            code = self._authorize(new_csrf_token, **kwargs)

        logging.debug("Authorization complete.") # log.INFO

        # This authorization code is passed to the API to get the access token & refresh token.
        return code




import json
import logging
import requests

from oauthlib.oauth2 import MismatchingRedirectURIError

from .status_codes import codes as withings_codes


# Authorizer Exceptions

class MismatchingRedirectURIError(MismatchingRedirectURIError):
    def __init__(self, redirect_uri=None,
                 description=None, uri=None, state=None,
                 status_code=None, request=None):

        self.redirect_uri = redirect_uri

        if redirect_uri:
            self.description = "CSRF Warning! The redirect URI provided ({}) is missing or does not match partner callback url." \
                .format(redirect_uri)
        else:
            self.description = "CSRF Warning! The redirect URI provided is missing or does not match partner callback url."

        super().__init__(description=None, uri=None, state=None,
            status_code=None, request=None)


# Client Exceptions

class StatusException(Exception):
    def __init__(self, status_code, error, *args, **kwargs):
        message = '({}) {}'.format(status_code, error)
        super().__init__(message, *args, **kwargs)


class AuthenticationFailedException(StatusException):
    '''Withings response status exception'''

class InvalidParamsException(StatusException):
    '''Withings response status exception'''
# from oauthlib.oauth2.rfc6749.errors import InvalidRequestFatalError # description, uri, state, status_code, request

class UnauthorizedException(StatusException):
    '''Withings response status exception'''
# from oauthlib.oauth2.rfc6749.errors import UnauthorizedClientError # description, uri, state, status_code, request

class ErrorOccurredException(StatusException):
    '''Withings response status exception'''

class TimeoutException(Exception):
    '''Withings response status exception'''
# from requests.exceptions import Timeout #(response (optional), request(even more optional, as it will attempt to take a request out of an included response.))

class BadStateException(StatusException):
    '''Withings response status exception'''

class TooManyRequestsException(StatusException):
    '''Withings response status exception'''
# from werkzeug.exceptions import TooManyRequests(description=None, response=None, retry_after=None)

class NotImplementedException(Exception):
    '''Withings response status exception'''
# NotImplementedError




# raise_for_status, raise_from_error, detect_and_raise_error, raise_exceptions
def raise_for_status(response):

    from functools import partial
    
    if isinstance(response, dict) and \
            'status' in response:
        status = response['status']
    elif isinstance(response, requests.models.Response):
        # requests.raise_for_status here first?

        try:
            response = json.loads(response.content.decode('utf8'))
            # logging.info(response)
            status = response['status']
        except ValueError:
            logging.error("Returned content isn't JSON format.")
            logging.error(response.__dict__)
            raise
    else:
        status = int(response.status_code)


    # if status_code in withings_codes.operation_was_successful:
    #     pass

    if status in withings_codes.authentication_failed:
        # {"status":401,"body":{},"error":"XRequestID: Not provided invalid_token: The access token provided is invalid"}
        raise AuthenticationFailedException(status, response['error'])

    elif status in withings_codes.invalid_params:
        raise InvalidParamsException(status, response['error'])
    elif status in withings_codes.unauthorized:
        raise UnauthorizedException(status, response['error'])
    elif status in withings_codes.an_error_occurred:
        raise ErrorOccurredException(status, response['error'])
    elif status in withings_codes.timeout:
        raise TimeoutException(status, response['error'])
    elif status in withings_codes.bad_state:
        raise BadStateException(status, response['error'])
    elif status in withings_codes.too_many_request:
        raise TooManyRequestsException(status, response['error'])
    elif status in withings_codes.not_implemented:
        raise NotImplementedException(status, response['error'])
    
    else:
        pass




if __name__ == "__main__":
    class Resp:
        status_code = "305"

    raise_for_status(Resp)


from enum import Enum
import numbers

# TODO: Probably move this class into structures.py
class ApiCodeEnum(Enum):
    '''Case-insensitive; unique; auto indexed; reversible'''

    def __new__(cls, value, code=None):
        obj = object.__new__(cls)
        obj._value_ = value
        
        # code defaults to index
        obj.code = code or len(cls.__members__)
        
        return obj


    def __init__(self, *args):
        '''Check that no duplicate NAME, VALUE, or CODE exists.'''

        cls = self.__class__
        
        if any(self.value.lower() == e.value.lower() for e in cls):
            raise ValueError(
                "Case-insensitive alias already exists in Enum: %s"
                % self.value.lower())
        
        # necessary case-insensitive check on name?
        elif any(self.name.lower()  == e.lower() for e in cls._member_names_):
            raise ValueError(
                "Case-insensitive name already exists in Enum: %s"
                % self.name.upper())

        if any(self.code == e.code for e in cls):
            raise ValueError(
                "Code already exists in Enum: %s"
                % self.code)


    @classmethod
    def _missing_(cls, value):
        '''cls._member_names_,  cls._value2member_map_,  cls._member_names_,
           cls.__members__ == cpy of cls._member_map_'''

        # Check if value is an INDEX
        if isinstance(value,int):
            code2member_map = {v.code:v for k,v in cls.__members__.items()}
            if value in code2member_map:
                return code2member_map[value]

        else:
            # Check for case-insensitive member VALUE
            upper_value2member = {k.upper():v for k,v in cls._value2member_map_.items()}
            if value.upper() in upper_value2member:
                return upper_value2member[value.upper()]

            # Check for case insensitive member NAME
            upper_members = {k.upper():v for k,v in cls.__members__.items()}
            if value.upper() in upper_members:
                return upper_members[value.upper()]

    def __int__(self):
        return self.code

    def __eq__(self, other):
        
        if isinstance(other, numbers.Real):
            return self.code == other
        else:
            return self.value == other



class MeasureType(ApiCodeEnum):
    WEIGHT = 'Weight',                          1   # (kg)
    HEIGHT = 'Height',                          4   # (meter)
    FAT_FREE_MASS = 'Fat Free Mass',            5   # (kg)
    BFP = 'Fat Ratio',                          6   # (%)
    FAT_MASS_WEIGHT = 'Fat Mass Weight',        8   # (kg)
    DIASTOLIC_BP = 'Diastolic Blood Pressure',  9   # (mmHg)
    SYSTOLIC_BP = 'Systolic Blood Pressure',   10   # (mmHg)
    PULSE = 'Heart Pulse',                     11   # (bpm) - only for BPM and scale', # devices
    TEMP = 'Temperature',                      12   # (celsius)
    SP02 = 'SP02',                             54   # (%)
    BODY_TEMP = 'Body Temperature',            71   # (celsius)
    SKIN_TEMP = 'Skin Temperature',            73   # (celsius)
    MUSCLE_MASS = 'Muscle Mass',               76   # (kg)
    HYDRATION = 'Hydration',                   77   # (kg)
    BONE_MASS = 'Bone Mass',                   88   # (kg)
    PULSE_WAVE_VEL = 'Pulse Wave Velocity',    91   # (m/s)
    V02_MAX = 'VO2 max',                      123   # Numerical measurement of your body’s ability to consume oxygen', # (ml/min/kg).


class SleepState(ApiCodeEnum):
    AWAKE = 'awake'    # 0
    LIGHT = 'light'    # 1
    DEEP = 'deep'      # 2
    REM = 'rem'        # 3



if __name__ == '__main__':
    print(int(MeasureType('Systolic Blood Pressure')))
    print(int(MeasureType('heart pulse')))
    print(MeasureType(10).value)
    print(MeasureType(MeasureType.BFP))


import pathlib
import re
from setuptools import setup

HERE = pathlib.Path(__file__).parent

# Get long_description value
README = (HERE/'README.md').read_text()

#Extract version value from __init__.py
version_regex = r'__version__ = ["\'](.+)["\']'
text = (HERE/"withings/__init__.py").read_text()
VERSION = re.search(version_regex, text).group(1)

setup(
    name="python-withings",
    version=VERSION,
    description="Wrapper package for Withings API",
    long_description_content_type="text/markdown",
    long_description=README,
    url="https://github.com/strohganoff/python-withings",
    packages=["withings", "withings/oauth2"],
    include_package_data=True,
    install_requires=["requests-oauthlib","python-dateutil"],
)